/*
  Name: Compiler.cpp
  Copyright: Polecatsoft 2005
  Game: Mercenary Soldier
  Author: Master Lenman
  Date: -
  Description: This is the compiler for Mercenary Soldier Script
*/

#include <fstream>
#include <string>
#include <sstream>
using namespace std;

#include "Compiler.h"

//Script types (used for compiler only)
const int
TILESCRIPT = 0,
UNITSCRIPT = 1;

enum parserVarTypes
{
	Vnone = 0, //No parameter
	Vread = 1, //Variable only parameter, read access
	Vwrite = 2,//Variable only parameter, write access
	Vnumber = 3,//Variable or constant, read access
	Vstring = 4 //Rest of line is a string
};

const int parse_functions = 22;
const char *parse_name[parse_functions] = {    "set",  "equal", "smaller", "inc", "dec", "add",  "sub",  "mul",  "div",  "quit", "quitn", "quita", "loadsound", "playsound", "deletesound", "stopsound", "msg",  "endl", "msgvar", "graphic", "goto",   "msgkey" };
const char parse_bytecode[parse_functions] =  { SET,    EQUAL,   SMALLER,   INC,   DEC,   ADD,    SUB,    MUL,    DIV,    QUIT,   QUITN,   QUITA,   LOADSOUND,   PLAYSOUND,    DELETESOUND,  STOPSOUND,   MSG,    ENDL,   MSGVAR,   GRAPHIC,   GOTO,     MSGKEY  };
const int parse_params[3][parse_functions] = {{ Vwrite, Vread,   Vnumber,   Vwrite,Vwrite,Vwrite, Vwrite, Vwrite, Vwrite, Vread,  Vread,   Vnone,   Vwrite,      Vread,        Vread,        Vread,       Vstring,Vnone,  Vread,    Vnumber,   Vstring,  Vnumber },
{												Vnumber,Vnumber, Vnumber,   Vnone, Vnone, Vnumber,Vnumber,Vnumber,Vnumber,Vnone,  Vnone,   Vnone,   Vstring,     Vnumber,      Vnone,        Vnone,       Vnone,  Vnone,  Vnone,    Vnumber,   Vnone,    Vnone   },
{                                               Vnone,  Vnone,   Vnone,     Vnone, Vnone, Vnone,  Vnone,  Vnone,  Vnone,  Vnone,  Vnone,   Vnone,   Vnone,       Vnumber,      Vnone,        Vnone,       Vnone,  Vnone,  Vnone,    Vnumber,   Vnone,    Vnone   }};

//This variable stores all constants
const int USEDPREDEFINEDCONSTS = 37;
const string STDCONSTS[USEDPREDEFINEDCONSTS] = { "ENEMYDEAD", "THISX", "THISY", "THISGRAPHIC", "DEBUGMODE", "THISID", "MAXLIFE", "MAXARMOR",
//Continue with definitions 8-15
"MISSIONSOLVED", "FALSE", "TRUE", "DOELSE", "ZFLAG", "ENEMIESLEFT", "", "",
//Continue with definitions 16-23
"POINTS1", "POINTS2", "OBJECTIVE1", "OBJECTIVE2", "OBJECTIVE3", "OBJECTIVE4", "OBJECTIVE5", "W1AMMO",
//Continue with definitions 24-31
"W2AMMO", "W3AMMO", "W4AMMO", "W5AMMO", "LIFE", "ARMOR", "HEALREFILLS", "AMMOREFILLS",
//Continue with definitions 32-36
"CALLER", "ACTIVEOBJECTIVE", "PISTOLOFF", "", "" };

string consts[ADDRESSRANGE];

//Transforms the string constant to an int, storing the result in retVal
//Returns false on error (e.g. constant is no number)
bool string2int(string constant, int &retVal);
//Extracts the number of the variable from the string constant, storing the result in retVal
//Returns false if constant contains neither a valid variable nor a const
bool const2int(string constant, int &retVal);

//Function that will log the error errString in logfile
//Always returns false
bool compileError(ofstream &logfile, string errString);

//This class stores the byte code produced by the compiler
class compilerCode
{
private:
	int size;
	int maxSize;
	char *data;

	void makeBigger();
public:
	compilerCode();
	~compilerCode();
	bool output(string filename);
	void write(const char *newData, int bytes);
	void read(const int readPos, char *output, int bytes)
	{
		for(int i = 0; i < bytes; ++i)
		{
			output[i] = data[readPos + i];
		}
	}
	char getLast(int back = 1);
	void backspace(int spaces = 1);
	int getPos();
};

//Reads a variable from readfile, using buffer as a buffer, outputting the result in outputCode
//Returns false on failure
//This function is useful if it's sure that only a variable is allowed
bool readVar(ifstream &readfile, string &buffer, compilerCode &outputCode, bool prot = false);

bool readVarOrNum(ifstream &readfile, string &buffer, compilerCode &outputCode);

//This function compiles the file inputPath and stores the result in the file outputPath
//Logs into compiler.log, returns false on error
bool compile(string inputPath, string outputPath, unsigned char *output, int *outputSize)
{
	//Check paths and open logfile and scriptfile
	ofstream logfile("compiler.log");
	logfile << "Reading from " << inputPath << " and writing to " << outputPath << ".\n\n";
	ifstream readfile(inputPath.c_str());
	if(readfile.fail())
		return compileError(logfile, "Failed to open file " + inputPath);

	//Clear output file
	if(outputPath.length())
	{
		ofstream clearFile(outputPath.c_str(), ios::binary);
		if(clearFile.fail())
			return compileError(logfile, "Failed to open file " + outputPath);
		char tempNull = 0;
		clearFile.write(&tempNull, 1);
		clearFile.close();
	}

	compilerCode compData;

	int i;
	bool backupLog = false;

	int ends = 0;
	int inEvent = 0;
	int onLoadUsed = 0, onUseUsed = 0;
	int onEnterUsed = 0, onLeaveUsed = 0, onExplodeUsed = 0, duringTouchUsed = 0;
	int onKillUsed = 0, onCallUsed = 0;
	int globalUsed = 0;
	int scripttype = -1;
	
	//For jumping operations
	const int MAXJMPS = 64;
	int jmpPos[MAXJMPS];
	string jmpName[MAXJMPS];
	int jmpCallPosition[MAXJMPS];
	int curJmpID = 0;
	for(i = 0; i < MAXJMPS; ++i)
	    jmpPos[i] = jmpCallPosition[i] = -1;

	logfile << "Beginning to read and compile.\n" << endl;
	string nextChar;

	for(i = 0; i < USEDPREDEFINEDCONSTS; ++i)
		consts[i] = STDCONSTS[i];
	for(; i < ADDRESSRANGE; ++i)
		consts[i].erase();

	while(readfile >> nextChar)
	{
		//Comments
		if(nextChar[0] == ';')
		{
			readfile.ignore(1024, '\n');
			continue;
		}
		//A new event
		if(nextChar == string("event"))
		{
			char data[2];
			data[0] = EVENT;
			readfile >> nextChar;
			if(nextChar == string("onLoad"))
			{
				if(onLoadUsed)
					return compileError(logfile, "The event onLoad is defined multiple times");
				data[1] = ONLOAD;
				onLoadUsed = true;
			}
			else if(nextChar == string("onUse"))
			{
				if(onUseUsed)
					return compileError(logfile, "The event onUse is defined multiple times");
				data[1] = ONUSE;
				onUseUsed = true;
			}
			else if(nextChar == string("onEnter"))
			{
				if(scripttype < 0)
					scripttype = TILESCRIPT;
				else if(scripttype == UNITSCRIPT)
					return compileError(logfile, "The event onEnter is illegal in a script for units");
				if(onEnterUsed)
					return compileError(logfile, "The event onEnter is defined multiple times");
				data[1] = ONENTER;
				onEnterUsed = true;
			}
			else if(nextChar == string("onLeave"))
			{
				if(scripttype < 0)
					scripttype = TILESCRIPT;
				else if(scripttype == UNITSCRIPT)
					return compileError(logfile, "The event onLeave is illegal in a script for units");
				if(onLeaveUsed)
					return compileError(logfile, "The event onLeave is defined multiple times");
				data[1] = ONLEAVE;
				onLeaveUsed = true;
			}
			else if(nextChar == string("onExplode"))
			{
				if(scripttype < 0)
					scripttype = TILESCRIPT;
				else if(scripttype == UNITSCRIPT)
					return compileError(logfile, "The event onExplode is illegal in a script for units");
				if(onExplodeUsed)
					return compileError(logfile, "The event onExplode is defined multiple times");
				data[1] = ONEXPLODE;
				onExplodeUsed = true;
			}
			else if(nextChar == string("duringTouch"))
			{
				if(scripttype < 0)
					scripttype = TILESCRIPT;
				else if(scripttype == UNITSCRIPT)
					return compileError(logfile, "The event duringTouch is illegal in a script for units");
				if(duringTouchUsed)
					return compileError(logfile, "The event duringTouch is defined multiple times");
				data[1] = DURINGTOUCH;
				duringTouchUsed = true;
			}
			else if(nextChar == string("onKill"))
			{
				if(scripttype < 0)
					scripttype = UNITSCRIPT;
				else if(scripttype == TILESCRIPT)
					return compileError(logfile, "The event onKill is illegal in a script for tiles");
				if(onKillUsed)
					return compileError(logfile, "The event onKill is defined multiple times");
				data[1] = ONKILL;
				onKillUsed = true;
			}
			else if(nextChar == string("onCall"))
			{
				if(scripttype < 0)
					scripttype = UNITSCRIPT;
				else if(scripttype == TILESCRIPT)
					return compileError(logfile, "The event onCall is illegal in a script for tiles");
				if(onCallUsed)
					return compileError(logfile, "The event onCall is defined multiple times");
				data[1] = ONCALL;
				onCallUsed = true;
			}
			else if(nextChar == string("global"))
			{
				if(globalUsed)
					return compileError(logfile, "The event global is defined multiple times");
				data[1] = GLOBAL;
				globalUsed = true;
			}
			else
				return compileError(logfile, "Unknown event " + nextChar);
			logfile << "Compiling event " << nextChar << "..." << endl;
			compData.write(data, 2);
			if(inEvent)
				compileError(logfile, "Illegal event " + nextChar + " inside of event");
			inEvent = data[1];
		}
		//End of if or end of event
		else if(nextChar == string("end"))
		{
			if(!inEvent)
				compileError(logfile, "Illegal end outside of event");
			char output;
			inEvent = 0;
			output = char(EVENTEND);
			if(compData.getLast(2) == EVENT)
			{
				compData.backspace(2);
				logfile << "WARNING: Empty event!" << endl;
				logfile << "Successfully compiled the event.\n" << endl;
				continue;
			}
			logfile << "Successfully compiled the event.\n" << endl;
			compData.write(&output, 1);
		}
		//end of an if or ifn statement
		else if(nextChar == string("endif"))
		{
			if(!inEvent)
				compileError(logfile, "Illegal endif outside of event");
			if(!ends)
				compileError(logfile, "Unexpected endif!");
			--ends;
			char output;
			output = char(IFEND);
			if(compData.getLast(2) == IF || compData.getLast(2) == IFN)
			{
				compData.backspace(2);
				logfile << "WARNING: Empty if or ifn instruction!" << endl;
				continue;
			}
			compData.write(&output, 1);
		}
		//jumping point declaration
		else if(nextChar[0] == ':')
		{
			if(!inEvent)
				return compileError(logfile, "Illegal jumping point outside of event");
			if(nextChar.length() == 1)
			    return compileError(logfile, "Empty ':' found!");
			nextChar = &nextChar[1];
			bool existed = false;
			int newEntryPos = 0;
			for(i = 0; i < MAXJMPS; ++i)
			{
				if(jmpName[i].empty() && jmpPos[i] == -1)
				{
					newEntryPos = i;
					break;
				}
				if(nextChar == jmpName[i])
				{
					existed = true;
					break;
				}
			}
			if(!existed)
			{
				jmpPos[newEntryPos] = compData.getPos();
				jmpName[newEntryPos] = compData.getPos();
			}
			else
			{
				if(jmpPos[i] == -1)
					jmpPos[newEntryPos] = compData.getPos();
				else
				    return compileError(logfile, "Multiple definition of jumping point " + jmpName[i]);
			}
		}
		//jmpif statement
		else if(nextChar == string("jmpif"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal jmpif outside of event");
			char data = char(JMPIF);
			compData.write(&data, 1);
			readfile >> nextChar;
			data = char(ISVAR);
			compData.write(&data, 1);
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);
			readfile >> nextChar;
			short jmpID = -1;
			bool existed = false;
			int newEntryPos = 0;
			for(i = 0; i < MAXJMPS; ++i)
			{
				if(jmpName[i].empty() && jmpPos[i] == -1)
				{
					newEntryPos = i;
					break;
				}
				if(nextChar == jmpName[i])
				{
					existed = true;
					break;
				}
			}
			if(!existed)
			{
				jmpName[newEntryPos] = nextChar;
				i = newEntryPos;
			}
			jmpID = i;
			jmpCallPosition[curJmpID++] = compData.getPos();
			compData.write((char*)(&jmpID), 2);
		}
		//jmpifn statement
		else if(nextChar == string("jmpifn"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal jmpifn outside of event");
			char data = char(JMPIFN);
			compData.write(&data, 1);
			readfile >> nextChar;
			data = char(ISVAR);
			compData.write(&data, 1);
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);
			readfile >> nextChar;
			short jmpID = -1;
			bool existed = false;
			int newEntryPos = 0;
			for(i = 0; i < MAXJMPS; ++i)
			{
				if(jmpName[i].empty() && jmpPos[i] == -1)
				{
					newEntryPos = i;
					break;
				}
				if(nextChar == jmpName[i])
				{
					existed = true;
					break;
				}
			}
			if(!existed)
			{
				jmpName[newEntryPos] = nextChar;
				i = newEntryPos;
			}
			jmpID = i;
			jmpCallPosition[curJmpID++] = compData.getPos();
			compData.write((char*)(&jmpID), 2);
		}
		//If statement
		else if(nextChar == string("if"))
		{
			//return compileError(logfile, "Use of deprecated function 'if'");
			logfile << "Warning! if is deprecated, use jmpifn instead!" << endl;
			if(!inEvent)
				return compileError(logfile, "Illegal if outside of event");
			++ends;
			char data = char(IF);
			compData.write(&data, 1);
			readfile >> nextChar;
			data = char(ISVAR);
			compData.write(&data, 1);
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			if(var == 9 || var == 10)
				logfile << "WARNING: Use of constant TRUE or FALSE in an if statement\n";
			compData.write(&data, 1);
		}
		//If not statement
		else if(nextChar == string("ifn"))
		{
			//return compileError(logfile, "Use of deprecated function 'ifn'");
			logfile << "Warning! if is deprecated, use jmpif instead!" << endl;
			if(!inEvent)
				return compileError(logfile, "Illegal ifn outside of event");
			++ends;
			char data = char(IFN);
			compData.write(&data, 1);
			readfile >> nextChar;
			data = char(ISVAR);
			compData.write(&data, 1);
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			if(var == 9 || var == 10)
				logfile << "WARNING: Use of constant TRUE or FALSE in an if statement\n";
			compData.write(&data, 1);
		}
		//Show ingame message and start new line
		else if(nextChar == string("msgendl"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal msgendl outside of event");
			readfile.get();
			getline(readfile, nextChar, '\n');
			char data = char(MSG);
			compData.write(&data, 1);
			compData.write(nextChar.c_str(), nextChar.length());
			data = 0;
			compData.write(&data, 1);
			data = char(ENDL);
			compData.write(&data, 1);
		}
		//Define new constant
		else if(nextChar == string("def"))
		{
			string varName;
			readfile >> varName >> nextChar;
			if(inEvent)
				return compileError(logfile, "Defining the constant " + nextChar + " in an event is illegal.");
			int temp;
			if(!const2int(varName, temp))
				return compileError(logfile, "Unknown variable name " + varName);
			for(i = 0; i < ADDRESSRANGE; ++i)
			{
				if(nextChar == consts[i])
					return compileError(logfile, "The constant " + nextChar + " is already defined");
			}
			consts[temp] = nextChar;
			if(temp < USEDPREDEFINEDCONSTS)
				logfile << "WARNING: Assigning new definition to constant variable!\n";
		}
		//Read defines from an external file
		else if(nextChar == string("include"))
		{
			readfile.get();
			getline(readfile, nextChar, '\n');
			if(inEvent)
				return compileError(logfile, "Including the file " + nextChar + " in an event is illegal");
			logfile << "Including external file " << nextChar << ".\n";

			ifstream incFile(nextChar.c_str());
			if(incFile.fail())
				return compileError(logfile, "Opening the file " + nextChar + " failed");
			while(incFile >> nextChar)
			{
				//Comments
				if(nextChar[0] == ';')
				{
					incFile.ignore(1024, '\n');
					continue;
				}
				//Define new constant
				else if(nextChar == string("def"))
				{
					string varName;
					incFile >> varName >> nextChar;
					int temp;
					if(!const2int(varName, temp))
						return compileError(logfile, "Unknown variable name " + varName);
					for(i = 0; i < ADDRESSRANGE; ++i)
					{
						if(nextChar == consts[i])
							return compileError(logfile, "The constant " + nextChar + " is already defined");
					}
					consts[temp] = nextChar;
					if(temp < USEDPREDEFINEDCONSTS)
						logfile << "WARNING: Assigning new definition to constant variable!\n";
				}
				else
					return compileError(logfile, "An included file must not contain a " + nextChar);
			}
			logfile << '\n';
			incFile.close();
		}
		//Show debug message
		else if(nextChar == string("debugMsg"))
		{
			readfile.get();
			getline(readfile, nextChar, '\n');
			logfile << "DEBUG MESSAGE: " << nextChar << endl;
		}
		//Copy content of compiler.log to a file named backup.log
		else if(nextChar == string("logbackup"))
		{
			backupLog = true;
		}
		//Unknown command
		else
		{
			for(i = 0; ; ++i)
			{
				if(i == parse_functions)
					return compileError(logfile, "Unknown command " + nextChar);
				if(nextChar == string(parse_name[i]))
				{
					if(!inEvent)
						return compileError(logfile, string("Illegal ") + nextChar + string(" outside of event"));

					char data = parse_bytecode[i];
					compData.write(&data, 1);
					int var;
					for(int par = 0; par < 3; ++par)
					{
						switch(parse_params[par][i])
						{
							//No further parameters
						case Vnone:
							par = 3;
							break;
							//Read access on variable
						case Vread:
							//Write access on variable
						case Vwrite:
							readfile >> nextChar;
							if(!const2int(nextChar, var))
								return compileError(logfile, "Unknown variable name " + nextChar);
							if(parse_params[i][par] == Vwrite && var < PROTECTEDRANGE)
								return compileError(logfile, "Access to " + nextChar + " denied");
							data = char(ISVAR);
							compData.write(&data, 1);
							data = char(var);
							compData.write(&data, 1);
							break;
							//Read access on variable or number
						case Vnumber:
							readfile >> nextChar;
							int var;
							if(string2int(nextChar, var))
							{
								data = ISNUM;
								compData.write(&data, 1);
								compData.write((char *)(&var), 4);
							}
							else if(const2int(nextChar, var))
							{
								data = ISVAR;
								compData.write(&data, 1);
								data = char(var);
								compData.write(&data, 1);
							}
							else
								return compileError(logfile, "Unknown variable name " + nextChar);
							break;
							//String
						case Vstring:
							readfile.get();
							getline(readfile, nextChar, '\n');
							compData.write(nextChar.c_str(), nextChar.length());
							data = 0;
							compData.write(&data, 1);

							par = 3;
							break;
						}
					}
					break;
				}
			}
		}
	}

	if(ends)
		return compileError(logfile, "Unexpected end of file");
	char data = char(SCRIPTEND);
	compData.write(&data, 1);
	for(i = 0; i < MAXJMPS; ++i)
	{
		if(jmpCallPosition[i] == -1)
		    break;
		short ID;
		compData.read(jmpCallPosition[i], (char*)(&ID), 2);
		//TODO: Finish this up
	}
	logfile << "Successfully finished compiling.\n\n";
	logfile << "Beginning to write compiled data.\n";
	if(outputPath.length())
	{
		if(!compData.output(outputPath))
			return compileError(logfile, "Outputting bytecode failed!");
	}
	if(output)
	{
		if(compData.getPos() > *outputSize)
		    return compileError(logfile, "Buffer for script data is not big enough");
		*outputSize = compData.getPos();
		compData.read(0, reinterpret_cast<char*>(output), *outputSize);
	}
	logfile << "Finished." << endl;
	logfile.close();

	if(backupLog)
	{
		ifstream logsource("compiler.log");
		ofstream logdestin("backup.log");
		string strBuf;
		while(!logsource.eof())
		{
			getline(logsource, strBuf);
			logdestin << strBuf << '\n';
		}
	}

	for(i = 0; i < ADDRESSRANGE; ++i)
		consts[i].erase();

	return true;
}

bool readVarOrNum(ifstream &readfile, string &buffer, compilerCode &outputCode)
{
	readfile >> buffer;
	int var;
	char data;
	if(string2int(buffer, var))
	{
		data = ISNUM;
		outputCode.write(&data, 1);
		outputCode.write((char *)(&var), 4);
	}
	else if(const2int(buffer, var))
	{
		data = ISVAR;
		outputCode.write(&data, 1);
		data = char(var);
		outputCode.write(&data, 1);
	}
	else
		return false;
	return true;
}

bool readVar(ifstream &readfile, string &buffer, compilerCode &outputCode, bool prot)
{
	readfile >> buffer;
	int var;
	char data;
	if(const2int(buffer, var))
	{
		data = char(var);
		outputCode.write(&data, 1);
	}
	else
		return false;
	if(prot && var < PROTECTEDRANGE)
		return false;
	return true;
}

bool string2int(string constant, int &retVal)
{
	istringstream str(constant);
	if(str >> retVal)
		return true;
	return false;
}

bool const2int(string constant, int &retVal)
{
	if(constant[0] == '%')
	{
		string temp = &constant[1];
		if(string2int(temp, retVal) && retVal < ADDRESSRANGE)
			return true;
		return false;
	}
	for(int i = 0; i < ADDRESSRANGE; ++i)
	{
		if(constant == consts[i])
		{
			retVal = i;
			return true;
		}
	}
	return false;
}

//Doubles the size of the code buffer
void compilerCode::makeBigger()
{
	char *temp = new char[maxSize << 1];
	maxSize <<= 1;
	for(int i = 0; i < size; ++i)
		temp[i] = data[i];
	delete [] data;
	data = temp;
}

//Create code buffer and set vars to default values
compilerCode::compilerCode() { data = new char[STARTSIZE]; maxSize = STARTSIZE; size = 0; }

//Frees code buffer
compilerCode::~compilerCode() { if(data) delete [] data; }

//Output the complete code into the file filename
bool compilerCode::output(string filename)
{
	ofstream outfile(filename.c_str(), ios::binary);
	if(outfile.fail())
		return false;
	outfile.write(data, size);
	//for(int i = 0; i < size; ++i)
	//	outfile.write(&data[i], 1);
	return true;
}

void compilerCode::write(const char *newData, int bytes)
{
	if(size + bytes >= maxSize - 1)
		makeBigger();
	for(int i = 0; i < bytes; ++i)
	{
		data[size + i] = newData[i];
	}
	size += bytes;
}

char compilerCode::getLast(int back)
{
	return data[size - back];
}

void compilerCode::backspace(int spaces)
{
	size -= spaces;
}

int compilerCode::getPos()
{
	return size;
}

bool compileError(ofstream &logfile, string errString)
{
	logfile << "FATAL ERROR: " << errString
		<< "\n\nAn error occured during compilation.\n"
		<< "Not writing compiled data.\n";
	logfile.close();
	return false;
}

//Function no longer needed!
//void constInit()
//{
//	//Set all definitions of variables that are not reserved to ""
//	for(int i = 32; i < ADDRESSRANGE; ++i)
//		consts[i] = "";
//}

//Old, bad parsing method
#if 0
/*
		//Set variable
		else if(nextChar == string("set"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal set outside of event");
			char data = char(SET);
			compData.write(&data, 1);
			int var;
			readfile >> nextChar;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			if(var < PROTECTEDRANGE)
				return compileError(logfile, "Access to " + nextChar + " denied");
			data = char(var);
			compData.write(&data, 1);

			readfile >> nextChar;
			if(string2int(nextChar, var))
			{
				data = ISNUM;
				compData.write(&data, 1);
				compData.write((char *)(&var), 4);
			}
			else if(const2int(nextChar, var))
			{
				data = ISVAR;
				if(var == compData.getLast())
				{
					compData.backspace(2);
					continue;
				}
				compData.write(&data, 1);
				data = char(var);
				compData.write(&data, 1);
			}
			else
				return compileError(logfile, "Unknown variable name " + nextChar);
		}
		//Compare 2 values if they're equal
		else if(nextChar == string("equal"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal equal outside of event");
			char data = char(EQUAL);
			compData.write(&data, 1);
			int var;
			readfile >> nextChar;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);

			readfile >> nextChar;
			if(string2int(nextChar, var))
			{
				data = ISNUM;
				compData.write(&data, 1);
				compData.write((char *)(&var), 4);
			}
			else if(const2int(nextChar, var))
			{
				data = ISVAR;
				compData.write(&data, 1);
				data = char(var);
				compData.write(&data, 1);
			}
			else
				return compileError(logfile, "Unknown variable name " + nextChar);
		}
		//Compare 2 values if the first one is smaller
		else if(nextChar == string("smaller"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal smaller outside of event");
			char data = char(SMALLER);
			compData.write(&data, 1);
			for(i = 0; i < 2; ++i)
			{
				readfile >> nextChar;
				int var;
				if(string2int(nextChar, var))
				{
					data = ISNUM;
					compData.write(&data, 1);
					compData.write((char *)(&var), 4);
				}
				else if(const2int(nextChar, var))
				{
					data = ISVAR;
					compData.write(&data, 1);
					data = char(var);
					compData.write(&data, 1);
				}
				else
					return compileError(logfile, "Unknown variable name " + nextChar);
			}
		}
		//Increment variable
		else if(nextChar == string("inc"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal inc outside of event");
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			if(var < PROTECTEDRANGE)
				return compileError(logfile, "Access to " + nextChar + " denied");
			char data = char(INC);
			compData.write(&data, 1);
			data = char(var);
			compData.write(&data, 1);
		}
		//Decrement variable
		else if(nextChar == string("dec"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal dec outside of event");
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			if(var < PROTECTEDRANGE)
				return compileError(logfile, "Access to " + nextChar + " denied");
			char data = char(DEC);
			compData.write(&data, 1);
			data = char(var);
			compData.write(&data, 1);
		}
		//Add value to variable
		else if(nextChar == string("add"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal add outside of event");
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			if(var < PROTECTEDRANGE)
				return compileError(logfile, "Access to " + nextChar + " denied");
			char data = char(ADD);
			compData.write(&data, 1);
			data = char(var);
			compData.write(&data, 1);
			readfile >> nextChar;
			if(string2int(nextChar, var))
			{
				data = ISNUM;
				compData.write(&data, 1);
				compData.write((char *)(&var), 4);
			}
			else if(const2int(nextChar, var))
			{
				data = ISVAR;
				compData.write(&data, 1);
				data = char(var);
				compData.write(&data, 1);
			}
			else
				return compileError(logfile, "Unknown variable name " + nextChar);
		}
		//Substract value from variable
		else if(nextChar == string("sub"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal sub outside of event");
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			if(var < PROTECTEDRANGE)
				return compileError(logfile, "Access to " + nextChar + " denied");
			char data = char(SUB);
			compData.write(&data, 1);
			data = char(var);
			compData.write(&data, 1);
			readfile >> nextChar;
			if(string2int(nextChar, var))
			{
				data = ISNUM;
				compData.write(&data, 1);
				compData.write((char *)(&var), 4);
			}
			else if(const2int(nextChar, var))
			{
				data = ISVAR;
				compData.write(&data, 1);
				data = char(var);
				compData.write(&data, 1);
			}
			else
				return compileError(logfile, "Unknown variable name " + nextChar);
		}
		//Multiplie variable with value
		else if(nextChar == string("mul"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal mul outside of event");
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			if(var < PROTECTEDRANGE)
				return compileError(logfile, "Access to " + nextChar + " denied");
			char data = char(MUL);
			compData.write(&data, 1);
			data = char(var);
			compData.write(&data, 1);
			readfile >> nextChar;
			if(string2int(nextChar, var))
			{
				data = ISNUM;
				compData.write(&data, 1);
				compData.write((char *)(&var), 4);
			}
			else if(const2int(nextChar, var))
			{
				data = ISVAR;
				compData.write(&data, 1);
				data = char(var);
				compData.write(&data, 1);
			}
			else
				return compileError(logfile, "Unknown variable name " + nextChar);
		}
		//Divide variable by value
		else if(nextChar == string("div"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal mul outside of event");
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			if(var < PROTECTEDRANGE)
				return compileError(logfile, "Access to " + nextChar + " denied");
			char data = char(DIV);
			compData.write(&data, 1);
			data = char(var);
			compData.write(&data, 1);
			readfile >> nextChar;
			if(string2int(nextChar, var))
			{
				data = ISNUM;
				compData.write(&data, 1);
				compData.write((char *)(&var), 4);
			}
			else if(const2int(nextChar, var))
			{
				data = ISVAR;
				compData.write(&data, 1);
				data = char(var);
				compData.write(&data, 1);
			}
			else
				return compileError(logfile, "Unknown variable name " + nextChar);
		}
		//quit if statement
		else if(nextChar == string("quit"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal quit outside of event");
			char data = char(QUIT);
			compData.write(&data, 1);
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);
		}
		//quit if not statement
		else if(nextChar == string("quitn"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal quitn outside of event");
			char data = char(QUITN);
			compData.write(&data, 1);
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);
		}
		//Quit statement
		else if(nextChar == string("quita"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal quita outside of event");
			char data = char(QUITA);
			compData.write(&data, 1);
		}
		//Load sound
		else if(nextChar == string("loadsound"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal loadsound outside of event");
			char data = char(LOADSOUND);
			compData.write(&data, 1);

			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);

			readfile.get();
			getline(readfile, nextChar, '\n');
			compData.write(nextChar.c_str(), nextChar.length());
			data = 0;
			compData.write(&data, 1);
		}
		//Destroy sound
		else if(nextChar == string("deletesound"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal deletesound outside of event");
			char data = char(DELETESOUND);
			compData.write(&data, 1);

			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);
		}
		//Play sound
		else if(nextChar == string("playsound"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal playsound outside of event");
			char data = char(PLAYSOUND);
			compData.write(&data, 1);

			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);

			for(i = 0; i < 2; ++i)
			{
				readfile >> nextChar;
				int var;
				if(string2int(nextChar, var))
				{
					data = ISNUM;
					compData.write(&data, 1);
					compData.write((char *)(&var), 4);
				}
				else if(const2int(nextChar, var))
				{
					data = ISVAR;
					compData.write(&data, 1);
					data = char(var);
					compData.write(&data, 1);
				}
				else
					return compileError(logfile, "Unknown variable name " + nextChar);
			}
		}
		//Stop playing a sound
		else if(nextChar == string("stopsound"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal stopsound outside of event");
			char data = char(STOPSOUND);
			compData.write(&data, 1);

			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			data = char(var);
			compData.write(&data, 1);
		}
		//Show ingame message
		else if(nextChar == string("msg"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal msg outside of event");
			readfile.get();
			getline(readfile, nextChar, '\n');
			char data = char(MSG);
			compData.write(&data, 1);
			compData.write(nextChar.c_str(), nextChar.length());
			data = 0;
			compData.write(&data, 1);
		}
		//Start new ingame line
		else if(nextChar == string("endl"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal endl outside of event");
			char data = char(ENDL);
			compData.write(&data, 1);
		}
		//Show variable ingame
		else if(nextChar == string("msgvar"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal msgvar outside of event");
			readfile >> nextChar;
			int var;
			if(!const2int(nextChar, var))
				return compileError(logfile, "Unknown variable name " + nextChar);
			char data = char(MSGVAR);
			compData.write(&data, 1);
			data = char(var);
			compData.write(&data, 1);
		}
		//Set new variable to tile
		else if(nextChar == string("graphic"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal graphic outside of event");
			if(inEvent == ONLOAD)
				return compileError(logfile, "Illegal graphic inside of onLoad event");
			char data = char(GRAPHIC);
			compData.write(&data, 1);

			//Get y pos on map
			//Get x pos on map
			//Get graphic to use
			for(i = 0; i < 3; ++i)
			{
				readfile >> nextChar;
				int var;
				if(string2int(nextChar, var))
				{
					data = ISNUM;
					compData.write(&data, 1);
					compData.write((char *)(&var), 4);
				}
				else if(const2int(nextChar, var))
				{
					data = ISVAR;
					compData.write(&data, 1);
					data = char(var);
					compData.write(&data, 1);
				}
				else
					return compileError(logfile, "Unknown variable name " + nextChar);
			}
		}
		//Goto next level
		else if(nextChar == string("goto"))
		{
			if(!inEvent)
				return compileError(logfile, "Illegal goto outside of event");
			char data = char(GOTO);
			compData.write(&data, 1);
			readfile >> nextChar;
			compData.write(nextChar.c_str(), nextChar.length());
			data = 0;
			compData.write(&data, 1);
		}*/
#endif //#if 0, bad parsing methods
