/*
  Name: ElementaryFunctions.cpp
  Copyright: Polecatsoft 2005
  Game: Mercenary Soldier
  Author: Master Lenman
  Date: -
  Description: This file contains basic functions which will
	  do things like setting Allegro up, showing the title screen
	  or blit the doubleBuffer to the screen.
*/

#include <allegro.h>
#include <libnet.h>
#include <almp3.h>
#include <string>
#include <sstream>
#include <fstream>
#include <ctime>
#include <cmath>
//#include <vector>
using namespace std;

#include "Sound.h"
#include "messageClass.h"
#include "keyStatesClass.h"
#include "ElementaryFunctions.h"
#include "Errors.h"
#include "Global.h"
#include "menues.h"
#include "requests.h"
#include "Entity.h"
#include "Unit.h"
#include "playerUnit.h"
#include "bytecode.h"
#include "Compiler.h"
#include "VM.h"
#include "mapClass.h"
#include "SPmapClass.h"
#include "bullet.h"
#include "SPgame.h"
#include "mpUnit.h"
#include "SMPgame.h"
#include "menues.h"
#include "mapEditor.h"
#include "CMPgame.h"
#include "mouse.h"

//Fix screen size to 800x600
const int	screenW = 800,
			screenH = 600;

const float SPEED[WEAPONTYPES] ={ 13.0, 13.0, 13.0, 15.0, 5.0, 9.5, 18.0, 18.0 };
const int DAMAGE[WEAPONTYPES] = {    5,   15,   10,   50,   0,   0,   20,   30 };
const int PIERCE[WEAPONTYPES] = {    0,    0,    0,    1,   0,   0,    2,    2 };
const int RATE[WEAPONTYPES] =   {   14,   20,    4,   32,  32,  64,    0,    0 };
const int BULLETS[WEAPONTYPES] ={    1,    5,    1,    1,   1,   1,    1,    1 };
const int SPREAD[WEAPONTYPES] = {    2,   10,    2,    0,   1,   2,    0,    0 };
const int RANGE[WEAPONTYPES] =  {   30,   14,   40,   60,  54,  55,    7,   10 };
const int SPLINTS[WEAPONTYPES] ={    0,    0,    0,    0,   7,   4,    0,    0 };
//const int AIWAIT[4] = {              5,    0,    3,    0    };
int realRange[WEAPONTYPES];

//Version 0.6 Alpha
const unsigned char
MAJOR_VERSION = 0,
MINOR_VERSION = 6,
VERSION_LETTER = 'A';
const string VERSIONSTRING = string("Mercenary Soldier 0.6 Alpha, built ") + string(__DATE__);

const char DELIMITER = 0x0D;

const int TIMER_FPS = 32;
const double PI = 3.14159265358979323846;

//timerCounterUpdater() will update timerCounter 32 times per second.
//This timer is used to make the game run at the same speed on all computers.
volatile int timerCounter = 0;
void timerCounterUpdater()
{
	++timerCounter;
	mouseCallback(0);
} END_OF_STATIC_FUNCTION(timerCounterUpdater)

//initialise allegro and some globals
void init()
{
	//initialise Allegro, timer, keyboard and sound
	allegro_init();

	if(install_timer() < 0)
		errors::timerError();
	if(install_keyboard() < 0)
		errors::keyboardError();
	if(install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL) < 0)
		errors::soundDriverError();
	if(install_mouse() < 0)
	    errors::mouseError();

	//Set a title for the window
	set_window_title("Mercenary Soldier - Polecatsoft");

	//load globals and set graphics mode
	global.init();

	clear(global.doublebuffer);
	textout_centre_ex(global.doublebuffer, global.size20, "Loading engine, please wait...", SCREEN_W >> 1, SCREEN_H >> 1, makecol(0xFF, 0xFF, 0xFF), -1);
	show();

    //Lock Timer Function and variable
	LOCK_FUNCTION(timerCounterUpdater);
	LOCK_VARIABLE(timerCounter);
	LOCK_FUNCTION(mouseCallback);

    //Set timer to 32 fps
	install_int_ex(timerCounterUpdater, BPS_TO_TIMER(TIMER_FPS));

	//Set Charset
	set_uformat(U_ASCII);

	//Initialise rand with the time and the randomly picked number 0xABCDE
	srand(time(0) * 0xABCDE);
}

void show(bool showMouse, BITMAP *bitmap)
{
	if(!bitmap)
	{
		bitmap = global.doublebuffer;
		if(keyStates[KEY_F9])
		{
			save_bitmap((string("screenshots/screenshot") + i2str(global.shotnumber++, 5) + string(".bmp")).c_str(), global.doublebuffer, NULL);
		}
	}
	if(showMouse)
		drawMouse();
	blit(bitmap, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
}

BITMAP *loadBitmap(string filename)
{
	BITMAP *retVal = load_bitmap(filename.c_str(), 0);
	if(!retVal)
		errors::loadBitError(filename);
	if(global.useVideoRAM)
	{
		BITMAP *buffer = create_video_bitmap(retVal->w, retVal->h);
		if(!buffer)
			errors::createError();
		blit(retVal, buffer, 0, 0, 0, 0, retVal->w, retVal->h);
		destroy_bitmap(retVal);
		retVal = buffer;
	}
	return retVal;
}

BITMAP *createBitmap(int width, int height)
{
	BITMAP *retVal = 0;
	if(global.useVideoRAM)
		retVal = create_video_bitmap(width, height);
	else
		retVal = create_bitmap(width, height);
	if(!retVal)
		errors::createError();
	return retVal;
}

BITMAP *createDatafileBitmap(DATAFILE *datafile, int index)
{
	BITMAP *original = (BITMAP*)datafile[index].dat;
	BITMAP *copy = createBitmap(original->w, original->h);
	blit(original, copy, 0, 0, 0, 0, original->w, original->h);
	return copy;
}

void setMode(int &mode)
{
	if(set_gfx_mode(mode, screenW, screenH, 0, 0) < 0)
	{
		mode ^= 3;
		if(set_gfx_mode(mode, screenW, screenH, 0, 0) < 0)
			errors::graphicModeError();
	}
}

void setMode(char &mode)
{
	int temp = mode;
	setMode(temp);
	mode = temp;
}

//The main() function is called first when the program is executed
int main(int argc, char *argv[])
{
	//If the first argument is "creator", create a map out of the file specified
	if(argc >= 2 && string(argv[1]) == string("creator"))
	{
		if(argc < 3)
			return 0;
		newMap(string(argv[2]));
		return 0;
	}
	//If the first argument is "compiler", compile the map specified
	if(argc >= 2 && string(argv[1]) == string("compiler"))
	{
		if(argc < 3)
			return 0;
		string inPath = argv[2], outPath;
		if(argc >= 4)
			outPath = argv[3];
		else
		{
			outPath = inPath;
			int length = outPath.length();
			outPath[length - 3] = 'm';
			outPath[length - 2] = 's';
			outPath[length - 1] = 's';
		}
		compile(inPath, outPath);
		return 0;
	}
	init();
	//If the first argument is "editor", launch the map editor
	if(argc >= 2 && string(argv[1]) == string("editor"))
	{
		install_mouse();
		//Use small font as default font in dialogs
		font = global.size12;

		string mapPath = "maps/";
		if(fileSelect("Choose map file to open.", mapPath, "msm", SCREEN_W, SCREEN_H))
		{
			mapEditor editor(mapPath);
			editor();
		}
		global.unload();
		remove_timer();
		return 0;
	}
#ifndef _DEBUG
	menu::intro();
#endif

	for(int i = 0; i < WEAPONTYPES; ++i)
		realRange[i] = int(float(RANGE[i]) * SPEED[i]);

	bool run = true;
	bool doMenu = true;
	bool restartGame = false;
	bool loadedGame = false;
	int menuReturn = 0;
	int healRefills = 0;
	int ammoRefills = 0;
	const int healRefillsPlus[3] = { 4, 2, 0 };
	const int ammoRefillsPlus[3] = { 3, 1, 0 };
	string mapPath;
	while(run && !restartGame)
	{
		if(doMenu)
			menuReturn = menu::main(menuReturn);
		else
			doMenu = true;
		switch(menuReturn)
		{
			//0 equals the New Game button
		case 0:
			{
				if(!loadedGame)
				{
					int newGameCatch = menu::newGameMenu();
					bool backToMain = false;
					switch(newGameCatch)
					{
					case 0:
						{
							//Load default map
							int difficulty = menu::getDifficulty();
							healRefills = healRefillsPlus[difficulty];
							ammoRefills = ammoRefillsPlus[difficulty];
						}
						if(menu::skipTutorial())
							mapPath.erase();
						else
						    mapPath = "tutorial";
						break;
					case 2:
						//Return to main menu
						backToMain = true;
						break;
					case 3:
						//Shooting range map
						mapPath = "range";
						break;
					case 4:
						//Tutorial map
						mapPath = "tutorial";
						break;
					}
					if(backToMain)
						break;
					if(mapPath.empty() || mapPath == "default")
					{
						ifstream defaultmap("maps/default");
						getline(defaultmap, mapPath);
						defaultmap.close();
					}
				}
				else
					loadedGame = false;

				while(true)
				{
					singlePlayerGame game(mapPath, healRefills, ammoRefills);
					int gameReturn = game(mapPath);
					//User requested abort
					if(gameReturn == 0)
					{
						mapPath.erase();
						break;
					}
					//Script requested abort
					if(gameReturn == 1)
					{
						//Game Over
						//Skip menu and show loading menu
						if(mapPath[0] == 'G' && mapPath[1] == 'O' && mapPath.length() == 2)
						{
							if(menu::savegameExists())
							{
								doMenu = false;
								menuReturn = 1;
							}
							else
								menuReturn = 0;
							//menu::gameOver();
							mapPath.erase();
							break;
						}
						//Player ran through the whole game,
						//Show end game screen
						else if(mapPath[0] == 'G' && mapPath[1] == 'S' && mapPath.length() == 2)
						{
							/* TODO (Lenman#1#): Add code for showing the end game screen */
							mapPath.erase();
							break;
						}
					}

					//Default behaviour,
					//Load next map
				}
			}
			break;

			//1 equals the Load Game button
		case 1:
			{
				int slot;
				slot = menu::savegameSelect(true);
				if(slot < 0)
					break;
				ifstream savegame(getSavegamePath(slot).c_str());
				getline(savegame, mapPath);
				savegame >> healRefills >> ammoRefills;
				savegame.close();
				loadedGame = true;
				doMenu = false;
				menuReturn = 0;
			}
			break;

			//2 equals the Multiplayer Button
		case 2:
			{
				int retCatch = menu::serverClient();
				if(retCatch == 1)
				{
					ifstream server("server.txt");
					string buf;
					server >> buf;
					server.close();
					clientMultiplayerGame game(buf.c_str());
					int retCatch = game();
				}
				else if(retCatch == 0)
				{
					serverMultiplayerGame game("testmap", 4, 2, false);
					int retCatch = game();
				}
			}
			break;

			//3 equals the Options Button
		case 3:
			if(menu::options(1))
			{
				restartGame = true;
//				if(restartGame)
//					allegro_message("You changed sensitive options,\nplease restart the game.");
			}
			break;

			//4 equals the exit button
		case 4:
			run = false;
			break;
		}
	}
	global.unload();
	if(restartGame)
		system((string("start \"\" \"") + string(argv[0]) + string("\"")).c_str());
	return 0;
} END_OF_MAIN()

void compileExtern(string scriptpath)
{
	string systemCommand = "scripts\\compiler.exe " + scriptpath + " scripts\\temp.mss";
	system(systemCommand.c_str());
}

void newMap(string scriptpath)
{
	ifstream scriptfile(scriptpath.c_str());
	if(scriptfile.fail())
		errors::fatalError("Failed to open file!");

	string mappath;
	getline(scriptfile, mappath);

	ofstream mapfile(mappath.c_str(), ios::binary);
	//Error checking for ofstreams is not necessary!
	//if(mapfile.fail())
	//	errors::fatalError("Failed to open file!");

	string tileset;
	getline(scriptfile, tileset);

	string charset;
	getline(scriptfile, charset);

	int height, width;
	char mapType;
	int newCampaign;
	scriptfile >> height >> width >> mapType >> newCampaign;
	scriptfile.ignore(128, '\n');
	string campaignName;
	if(newCampaign)
		getline(scriptfile, campaignName);

	mapfile.write((char *)&MAJOR_VERSION, 1);
	mapfile.write((char *)&MINOR_VERSION, 1);
	mapfile.write((char *)&VERSION_LETTER, 1);
	mapfile.write(&mapType, 1);
	mapfile.write(tileset.c_str(), tileset.length());
	mapfile.write(&DELIMITER, 1);

	mapfile.write(i2ch(height), sizeof(int));
	mapfile.write(i2ch(width), sizeof(int));
	int i;
	const char defaultContent[4] = { 0, 0, 0, 0 };
	for(i = 0; i < height * width; ++i)
		mapfile.write(defaultContent, 2);

	char buf = 0;
	mapfile.write(&buf, 1);
	buf = newCampaign & 1;
	mapfile.write(&buf, 1);
	if(newCampaign)
	{
		mapfile.write(campaignName.c_str(), campaignName.length());
		mapfile.write(&DELIMITER, 1);
	}

	mapfile.write(charset.c_str(), charset.length());
	mapfile.write(&DELIMITER, 1);
	mapfile.write(defaultContent, 4);
}

bool fileSelect(string message, string &path, string ext, int w, int h)
{
	char pathBuffer[1024];
	strncpy(pathBuffer, path.c_str(), 1024);
	int retVal = file_select_ex(message.c_str(), pathBuffer, ext.c_str(), 1024, w, h);
	if(!retVal)
		return false;
	path = pathBuffer;
	return true;
}

bool fileExists(string filename)
{
	ifstream testStream(filename.c_str());
	bool result = !testStream.fail();
	testStream.close();

	return result;
}

string i2str(int i, int minLength)
{
	ostringstream oss; // create a stream
	oss << i; // insert value to stream
	if((minLength -= oss.str().length()) <= 0)
		return oss.str(); // extract value and return it
	string retVal;
	while(retVal.length() < minLength)
		retVal.append(1, '0');
	return retVal + oss.str();
}

string getSavegamePath(int n)
{
	return (string("save/slot") + i2str(n) + string(".sav"));
}
