/*
  Name: Sound.cpp
  Copyright: Polecatsoft 2005
  Game: Mercenary Soldier
  Author: Master Lenman
  Date: -
  Description: This class provides a high level API for
	  loading and playing sounds of different formats
*/

#include <allegro.h>
#include <almp3.h>
#include <libnet.h>
#include <string>
#include <fstream>
using namespace std;

#include "Errors.h"
#include "Global.h"
#include "Sound.h"

enum activeStates
{
	INACTIVE = 0,
	SAMPLEACTIVE = 1,
	MIDIACTIVE = 2,
	MP3ACTIVE = 3,
};

Sound::Sound()
{
	sample = NULL;
	midi = NULL;
	mp3 = NULL;
	mp3Data = NULL;
	active = INACTIVE;
	soundEffect = true;
	autoPoll = false;
}

Sound::Sound(string filename, bool doPoll)
{
	Sound();
	load(filename, doPoll);
}

Sound::~Sound()
{
	destroy();
}

void Sound::load(string filename, bool doPoll)
{
	destroy();
	//Load wav file
	if(
		filename[filename.length() - 3] == 'w' &&
		filename[filename.length() - 2] == 'a' &&
		filename[filename.length() - 1] == 'v')
	{
		sample = load_sample(filename.c_str());
		if(!sample)
			errors::sampleLoadError(filename);

		active = SAMPLEACTIVE;
	}
	//Load mp3
	else if(
		filename[filename.length() - 3] == 'm' &&
		filename[filename.length() - 2] == 'p' &&
		filename[filename.length() - 1] == '3')
	{
		int length = file_size(filename.c_str());
		if(!length)
			errors::sampleLoadError(filename);
		mp3Data = new char[length];

		//ifstream mp3file(filename.c_str());
		//mp3file.read(mp3Data, length);
		//mp3file.close();
		PACKFILE *mp3readfile = pack_fopen(filename.c_str(), "r");
		pack_fread(mp3Data, length, mp3readfile);
		pack_fclose(mp3readfile);


		mp3 = almp3_create_mp3(mp3Data, length);
		if(!mp3)
			errors::sampleLoadError(filename);
		if(doPoll)
		{
			almp3_start_autopoll_mp3(mp3, 100);
			autoPoll = true;
		}
		else
		{
			autoPoll = false;
		}
		active = MP3ACTIVE;
	}
	//Load midi
	else if(
		(
			filename[filename.length() - 3] == 'm' &&
			filename[filename.length() - 2] == 'i' &&
			filename[filename.length() - 1] == 'd'
		)
		||
		(
			filename[filename.length() - 4] == 'm' &&
			filename[filename.length() - 3] == 'i' &&
			filename[filename.length() - 2] == 'd' &&
			filename[filename.length() - 1] == 'i'
		))
	{
		midi = load_midi(filename.c_str());
		if(!midi)
			errors::sampleLoadError(filename);

		active = MIDIACTIVE;
	}
	else
		errors::sampleLoadError(filename);
}

void Sound::play(int loop, int vol, int pan, int freq)
{
	if(active == MIDIACTIVE)
	{
		play_midi(midi, loop);
	}
	else if(active == SAMPLEACTIVE)
	{
		play_sample(sample, vol, pan, freq, loop);
	}
	else if(active == MP3ACTIVE)
	{
		almp3_play_ex_mp3(mp3, 24576, vol, pan, freq, loop);
	}

	this->loop = loop;
	this->vol = vol;
	this->pan = pan;
	this->freq = freq;
	soundEffect = !(vol == global.musicVol);
}

void Sound::stop()
{
	if(active == SAMPLEACTIVE)
	{
		stop_sample(sample);
	}
	else if(active == MIDIACTIVE)
	{
		stop_midi();
	}
	else if(active == MP3ACTIVE)
	{
		almp3_stop_mp3(mp3);
	}

	soundEffect = true;
}

void Sound::destroy()
{
	if(sample)
	{
		destroy_sample(sample);
		sample = NULL;
	}
	if(midi)
	{
		destroy_midi(midi);
		midi = NULL;
	}
	if(mp3)
	{
		if(autoPoll)
		{
			almp3_stop_autopoll_mp3(mp3);
			autoPoll = true;
		}
		autoPoll = false;

		almp3_destroy_mp3(mp3);
		mp3 = NULL;
		delete [] mp3Data;
		mp3Data = NULL;
	}
	active = INACTIVE;

	soundEffect = true;
}

int Sound::isActive()
{
	return active;
}

void Sound::poll()
{
	if(active == MP3ACTIVE && !autoPoll)
	{
		almp3_poll_mp3(mp3);
	}
}

void Sound::musicUpdate()
{
	if(!soundEffect)
	{
		if(active == SAMPLEACTIVE)
		{
			adjust_sample(sample, global.musicVol, pan, freq, loop);
			vol = global.musicVol;
		}
		else if(active == MP3ACTIVE)
		{
			almp3_adjust_mp3(mp3, global.musicVol, pan, freq, loop);
			vol = global.musicVol;
		}
	}
	else
	{
		if(active == SAMPLEACTIVE)
		{
			adjust_sample(sample, global.sfxVol, pan, freq, loop);
			vol = global.sfxVol;
		}
		else if(active == MP3ACTIVE)
		{
			almp3_adjust_mp3(mp3, global.sfxVol, pan, freq, loop);
			vol = global.sfxVol;
		}
	}
}
