/*
  Name: mapClass.cpp
  Copyright: Polecatsoft 2005
  Game: Mercenary Soldier
  Author: Master Lenman
  Date: -
  Description: This is the map class used in MP mode,
    it's the base class for the SP mode map class
*/

#include <allegro.h>
#include <almp3.h>
#include <libnet.h>
#include <string>
#include <fstream>
//#include <vector>
using namespace std;

#include "Sound.h"
#include "messageClass.h"
#include "ElementaryFunctions.h"
#include "requests.h"
#include "Entity.h"
#include "Unit.h"
#include "bytecode.h"
#include "Compiler.h"
#include "VM.h"
#include "Global.h"
#include "mapClass.h"
#include "Errors.h"

const int
TILESIZE = 32;

const int RADIUS = 8;

char mapClass::loadMap(ifstream &readfile, string filename)
{
	//Read header
	char retVal;
	char buffer[3];
	readfile.read(buffer, 3);
	if(buffer[0] != MAJOR_VERSION || buffer[1] != MINOR_VERSION || buffer[2] != VERSION_LETTER)
		errors::mapLoadError(filename, "Version mismatch");
	readfile.read(&retVal, 1);
	if(retVal == 'S')
		errors::mapLoadError(filename, "Wrong map type");
	mapType = retVal;

	//Read tileset
	string strBuffer;
	getline(readfile, strBuffer, DELIMITER);
	setTileset(strBuffer);
	animCountdown = 16;

	//Read size
	readfile.read(i2ch(width), sizeof(width));
	readfile.read(i2ch(height), sizeof(height));
	size = width * height;
	graphicData = new char[size];

	//Initialise VM for graphic command
	//VM.mapSetup(width, height, graphicData);

	//Create holders for bytecode
	//onEnter = new bytecode*[size];
	//onLeave = new bytecode*[size];
	//onUse = new bytecode*[size];
	//duringTouch = new bytecode*[size];
	//onExplode = new bytecode*[size];

	//string lastFive[5];

	//Load map data and scripts
	for(int i = 0; i < size; ++i)
	{
		readfile.read(&graphicData[i], 1);
		//Load script
		readfile.read(buffer, 1);

		//if(buffer)
		//   errors::mapLoadError(filename, "Scripts are not allowed on multiplayer maps!");
	}
	visibleTilesX = SCREEN_W / TILESIZE + 1;
	if(visibleTilesX > width)
		visibleTilesX = width;
	visibleTilesY = SCREEN_H / TILESIZE + 1;
	if(visibleTilesY > height)
		visibleTilesY = height;
	maxX = width * TILESIZE - SCREEN_W / 2;
	maxY = height * TILESIZE - SCREEN_H / 2;
}

void mapClass::setTileset(string filename)
{
	DATAFILE *datafile = load_datafile(filename.c_str());
	if(!datafile)
		errors::datafileError(filename);
	tileset = createDatafileBitmap(datafile, 1);
	if(tileset->w % TILESIZE)
		errors::tilesetLoadError(filename, "Each tile has to be 32 pixels wide");
	tileTypes = tileset->w / TILESIZE;
	hightData = new char[tileTypes];
	followUpData = new char[tileTypes];
	int i;
	for(i = 0; i < tileTypes; ++i)
	{
		hightData[i] = *((char*)datafile[0].dat + i);
	}
	for(i = 0; i < tileTypes; ++i)
	{
		followUpData[i] = *((char*)datafile[0].dat + i + tileTypes);
	}
	unload_datafile(datafile);
}

mapClass::~mapClass()
{
	if(graphicData)
		delete [] graphicData;
	if(hightData)
		delete [] hightData;

	if(tileset)
		destroy_bitmap(tileset);
}

void mapClass::draw(int &posX, int &posY)
{
	int positionX = MID(SCREEN_W >> 1, posX, maxX);
	int positionY = MID(SCREEN_H >> 1, posY, maxY);
	int minusX = (positionX - (SCREEN_W >> 1)) & (TILESIZE - 1);
	int minusY = (positionY - (SCREEN_H >> 1)) & (TILESIZE - 1);
	int balance = 0;
	if(posX >= (SCREEN_W >>1) + TILESIZE)
		balance = -1;

	int tileX = MID(0, (positionX - minusX) / TILESIZE - visibleTilesX / 2, maxX / TILESIZE) - balance;
	int tileY = MID(0, (positionY - minusY) / TILESIZE - visibleTilesY / 2, maxY / TILESIZE);

	int pos = 0;

	int y;
	for(y = 0; y < visibleTilesY + 1; ++y)
	{
		pos = getOffset(tileX, y + tileY);
		int x;
		for(x = 0; x < visibleTilesX; ++x)
		{
			drawTile(x * TILESIZE - minusX, y * TILESIZE - minusY, graphicData[pos], TILESIZE);
			++pos;
		}
	}
	posX = positionX - (SCREEN_W >> 1);
	posY = positionY - (SCREEN_H >> 1);//tileY * TILESIZE - minusY;
}

int mapClass::getMapHeight()
{
	return TILESIZE * height;
}

int mapClass::getMapWidth()
{
	return TILESIZE * width;
}

bool mapClass::handleMoveRequest(float &newPosX, float &newPosY, float oldPosX, float oldPosY)
{
	bool retVal = true;
	int oldTileX = int(oldPosX) >> 5,
		oldTileY = int(oldPosY) >> 5;

	int newTileX = int(newPosX) >> 5,
		newTileY = int(newPosY) >> 5;

	//Decline the request if the object wants to leave the map
	if(newPosX - RADIUS < 0.0 || newPosX + RADIUS > float(width * TILESIZE))
	{
		newPosX = oldPosX;
		retVal = false;
	}
	if(newPosY - RADIUS < 0.0 || newPosY + RADIUS > float(height * TILESIZE))
	{
		newPosY = oldPosY;
		retVal = false;
	}

	//Decline the request if the object's rights are equal or lower to the new tile's rights
	if(RIGHT_LOW <= hightData[graphicData[getOffset(newTileX, newTileY)]])
	{
		newPosX = oldPosX;
		newPosY = oldPosY;
		retVal = false;
	}
	return retVal;
}

char *mapClass::getGraphicDataAddress()
{
	return graphicData;
}

void mapClass::xyFromOffset(int &x, int &y, int offset)
{
	bool found = false;
	for(x = 0; x < width && !found; ++x)
	{
		for(y = 0; y < height && !found; ++y, found = (getOffset(x, y) == offset))
		{
			//if(getOffset(x, y) == offset)
			//{
			//	found = true;
			//	break;
			//}
		}
	}
}

void mapClass::poll()
{
	--animCountdown;
	if(!animCountdown)
	{
		animCountdown = 16;
		for(int i = 0; i < size; ++i)
			graphicData[i] = followUpData[graphicData[i]];
	}
}

const int mapClass::getTilesize()
{
	return TILESIZE;
}

/*nodeList mapClass::getPath(int startX, int startY, int endX, int endY)
{
	node start(startX, startY);
	node end(endX, endY);
	node current = start;
	nodeList retPath;
	while(!(current == end))
	{
		int distX = endX - current.getX();
		int distY = endY - current.getY();

		if(distY > distX && RIGHT_LOW >= hightData[graphicData[getOffset(current.getX(), current.getY())]])
		{
			//node temp(
		}
	}
	return retPath;
}*/
